!
! Program to solve a finite difference
! discretization of the screened Poisson equation:
! (d2/dx2)u + (d2/dy2)u - alpha u = f
! with zero Dirichlet boundary condition using the iterative
! Jacobi method with overrelaxation.
!
! RHS (source) function
!   f(x,y) = -alpha*(1-x^2)(1-y^2)-2*[(1-x^2)+(1-y^2)]
!
! Analytical solution to the PDE
!   u(x,y) = (1-x^2)(1-y^2)
!
! Current Version: Christian Iwainsky, RWTH Aachen University
! MPI C Version: Christian Terboven, RWTH Aachen University, 2006
! MPI Fortran Version: Dieter an Mey, RWTH Aachen University, 1999 - 2005
! Modified: Sanjiv Shah,        Kuck and Associates, Inc. (KAI), 1998
! Author:   Joseph Robicheaux,  Kuck and Associates, Inc. (KAI), 1998
!
! Input : n     - grid dimension in x direction
!         m     - grid dimension in y direction
!         alpha - constant (always greater than 0.0)
!         tol   - error tolerance for the iterative solver
!         relax - Successice Overrelaxation parameter
!         mits  - maximum iterations for the iterative solver
!
! On output
!       : u(n,m)       - Dependent variable (solution)
!       : f(n,m,alpha) - Right hand side function
!

module solver
contains

    !
    ! Performs one iteraton of the Jacobi method and computes
    ! the residual value.
    !
    ! NOTE: u(0,:), u(maxXCount,:), u(:,0) and u(:,maxYCount)
    ! are BOUNDARIES and therefore not part of the solution.
    !
    subroutine one_jacobi_iteration(xStart, yStart, maxXCount, maxYCount, &
        src, dst, deltaX, deltaY, alpha, omega, error)
        implicit none
        real(kind=8), intent(in) :: xStart, yStart
        integer, intent(in) :: maxXCount, maxYCount
        real(kind=8), dimension(0:maxXCount-1,0:maxYCount-1), intent(in) :: src
        real(kind=8), dimension(0:maxXCount-1,0:maxYCount-1), intent(out) :: dst
        real(kind=8), intent(in) :: deltaX, deltaY, alpha, omega
        real(kind=8), intent(out) :: error

        integer :: x, y
        real(kind=8) :: f, fX, fY
        real(kind=8) :: updateVal
        ! Coefficients
        real(kind=8) :: cx, cy, cc
        cx = 1.0_8/(deltaX**2)
        cy = 1.0_8/(deltaY**2)
        cc = -2.0_8*cx-2.0_8*cy-alpha

        error = 0.0

        do y = 1, maxYCount-2
            fY = yStart + (y-1)*deltaY
            do x = 1, maxXCount-2
                fX = xStart + (x-1)*deltaX
                f = -alpha*(1.0-fX*fX)*(1.0-fY*fY)-2.0*(1.0-fX*fX)-2.0*(1.0-fY*fY)
                updateVal = ((src(x-1,y)+src(x+1,y))*cx + &
                    (src(x,y-1)+src(x,y+1))*cy + &
                    src(x,y)*cc - f)/cc
                dst(x,y) = src(x,y) - omega*updateVal
                error = error + updateVal*updateVal
            end do
        end do
    end subroutine one_jacobi_iteration

    !
    ! Checks the error between numerical and exact solutions
    !
    subroutine checkSolution(xStart, yStart, maxXCount, maxYCount, u, &
        deltaX, deltaY, alpha, error)
        implicit none
        real(kind=8), intent(in) :: xStart, yStart
        integer, intent(in) :: maxXCount, maxYCount
        real(kind=8), dimension(0:maxXCount-1,0:maxYCount-1), intent(in) :: u
        real(kind=8), intent(in) :: deltaX, deltaY, alpha
        real(kind=8), intent(out) :: error

        integer :: x, y
        real(kind=8) :: fX, fY
        real(kind=8) :: localError

        error = 0.0

        do y = 1, maxYCount-2
            fY = yStart + (y-1)*deltaY
            do x = 1, maxXCount-2
                fX = xStart + (x-1)*deltaX
                localError = u(x,y) - (1.0-fX**2)*(1.0-fY**2)
                error = error + localError**2
            end do
        end do
    end subroutine checkSolution

    !
    ! Performs halo swap
    !
    subroutine swap_halos(maxXCount, maxYCount, src, prevRank, nextRank)
        use mpi
        implicit none
        integer, intent(in) :: maxXCount, maxYCount
        real(kind=8), dimension(0:maxXCount-1,0:maxYCount-1), intent(inout) :: src
        integer, intent(in) :: prevRank, nextRank

        integer :: ierr

        call MPI_SENDRECV(src(0,maxYCount-2), maxXCount, MPI_REAL8, nextRank, 0, &
            src(0,0), maxXCount, MPI_REAL8, prevRank, 0, &
            MPI_COMM_WORLD, MPI_STATUS_IGNORE, ierr)
        call MPI_SENDRECV(src(0,1), maxXCount, MPI_REAL8, prevRank, 0, &
            src(0,maxYCount-1), maxXCount, MPI_REAL8, nextRank, 0, &
            MPI_COMM_WORLD, MPI_STATUS_IGNORE, ierr)
    end subroutine swap_halos

end module solver

!
! Main program
!
program jacobi
    use solver
    use mpi
    implicit none
    integer :: n, m, mits
    real(kind=8) :: alpha, tol, relax
    real(kind=8) :: maxAcceptableError
    real(kind=8) :: error, localError, absoluteError
    real(kind=8), dimension(:,:), allocatable, target :: u1, u2
    real(kind=8), dimension(:,:), pointer :: u, u_old, tmp
    real(kind=8), dimension(:,:), allocatable :: u_big
    integer :: iterationCount, maxIterationCount

    ! Solution space is [-1, 1] x [-1, 1]
    real(kind=8), parameter :: xLeft = -1.0, xRight = 1.0
    real(kind=8), parameter :: yBottom = -1.0, yUp = 1.0

    real(kind=8) :: deltaX, deltaY

    ! MPI related variables
    integer :: myRank, numProcs, ierr
    integer :: prevRank, nextRank

    integer :: myStartRow, myRows
    real(kind=8) :: myYBottom

    integer :: rank, startRow, rows
    integer, dimension(:), allocatable :: counts, displs

    call MPI_INIT(ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD, myRank, ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, numProcs, ierr)

    prevRank = myRank - 1
    if (prevRank < 0) prevRank = MPI_PROC_NULL
    nextRank = myRank + 1
    if (nextRank >= numProcs) nextRank = MPI_PROC_NULL

    if (myRank == 0) then
        print *,"Input n m - grid dimension in x and y direction:"
        read *,n,m
        print *,"Input alpha - Helmholtz constant:"
        read *,alpha
        print *,"Input relax - successive over-relaxation parameter:"
        read *,relax
        print *,"Input tol - error tolerance for the iterative solver:"
        read *,tol
        print *,"Input mits - maximum solver iterations:"
        read *,mits
        print *,"->",n,m,alpha,relax,tol,mits
    end if

    ! Broadcast configuration data to all processes
    call MPI_BCAST(n, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    call MPI_BCAST(m, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)
    call MPI_BCAST(alpha, 1, MPI_REAL8, 0, MPI_COMM_WORLD, ierr)
    call MPI_BCAST(relax, 1, MPI_REAL8, 0, MPI_COMM_WORLD, ierr)
    call MPI_BCAST(tol, 1, MPI_REAL8, 0, MPI_COMM_WORLD, ierr)
    call MPI_BCAST(mits, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)

    myStartRow = (myRank * m) / numProcs
    myRows = ((myRank+1) * m) / numProcs - myStartRow

    allocate(counts(0:numProcs-1))
    allocate(displs(0:numProcs-1))

    do rank = 0, numProcs-1
        startRow = (rank * m) / numProcs
        rows = ((rank+1) * m) / numProcs - startRow

        displs(rank) = startRow*(n+2)
        counts(rank) = rows*(n+2)

        if (myRank == rank) then
            myStartRow = startRow
            myRows = rows
        end if
    end do

    allocate(u1(0:n+1,0:myRows+1))
    u_old => u1
    allocate(u2(0:n+1,0:myRows+1))
    u => u2
    if (.not.allocated(u1) .or. .not.allocated(u2)) then
        print '("[",I0,"] Not enough memory for two ",I0,"x",I0," matrices")', &
            myRank, n+2, myRows+2
        call MPI_ABORT(MPI_COMM_WORLD, 1, ierr)
        stop 1
    end if
    ! Zero the elements (including the boundaries)
    u_old = 0.
    u = 0.

    ! Master process also allocates an array that holds the entire solution
    if (myRank == 0) then
        allocate(u_big(0:n+1,0:m+1))
        if (.not.allocated(u_big)) then
            print *,"Not enough memory for the whole solution in master process"
            call MPI_ABORT(MPI_COMM_WORLD, 1, ierr)
            stop 1
        end if
    end if

    maxIterationCount = mits
    maxAcceptableError = tol

    deltaX = (xRight-xLeft)/(n-1)
    deltaY = (yUp-yBottom)/(m-1)

    myYBottom = yBottom + myStartRow*deltaY

    ! Distribute the initial solution approximation to all processes
    if (myRank == 0) then
        call MPI_SCATTERV(u_big(0,1), counts, displs, MPI_REAL8, &
            u_old(0,1), myRows*(n+2), MPI_REAL8, &
            0, MPI_COMM_WORLD, ierr)
    else
        ! u_big is not allocated in ranks other than 0 and the above reference
        ! to u_big(0,1) throws a run-time error if bounds checking is enabled
        call MPI_SCATTERV(u_big, counts, displs, MPI_REAL8, &
            u_old(0,1), myRows*(n+2), MPI_REAL8, &
            0, MPI_COMM_WORLD, ierr)
    end if

    iterationCount = 0
    error = 1.e+100_8

    ! Iterate as long as it takes to meet the convergence criterion
    do while (iterationCount < maxIterationCount .and. &
        error > maxAcceptableError)
        if (myRank == 0) &
            print *,"Iteration",iterationCount
        call swap_halos(n+2, myRows+2, u_old, prevRank, nextRank)
        call one_jacobi_iteration(xLeft, myYBottom, n+2, myRows+2, u_old, u, &
            deltaX, deltaY, alpha, relax, localError)
        ! Compute global error value
        call MPI_ALLREDUCE(localError, error, 1, MPI_REAL8, MPI_SUM, &
            MPI_COMM_WORLD, ierr)
        error = sqrt(error)/(n*m)
        if (myRank == 0) &
            print *,"   Error",error
        iterationCount = iterationCount+1
        ! Swap the buffers
        tmp => u_old
        u_old => u
        u => tmp
    end do
    if (myRank == 0) &
        print *,"Residual",error

    ! u_old points to the solution after the most recent buffers swap
    call checkSolution(xLeft, myYBottom, n+2, myRows+2, u_old, &
        deltaX, deltaY, alpha, localError)
    call MPI_ALLREDUCE(localError, absoluteError, 1, MPI_REAL8, MPI_SUM, &
        MPI_COMM_WORLD, ierr)
    absoluteError = sqrt(absoluteError)/(n*m)
    if (myRank == 0) &
        print *,"The error of the iterative solution is", absoluteError

    ! Gather back the whole solution
    if (myRank == 0) then
        call MPI_GATHERV(u_old(0,1), myRows*(n+2), MPI_REAL8, &
            u_big(0,1), counts, displs, MPI_REAL8, &
            0, MPI_COMM_WORLD, ierr)
    else
        ! See the comment about MPI_SCATTER invocation
        call MPI_GATHERV(u_old(0,1), myRows*(n+2), MPI_REAL8, &
            u_big, counts, displs, MPI_REAL8, &
            0, MPI_COMM_WORLD, ierr)
    end if

    ! Check the solution again to compare with the reduction result
    if (myRank == 0) then
        call checkSolution(xLeft, yBottom, n+2, m+2, u_big, &
            deltaX, deltaY, alpha, absoluteError)
        absoluteError = sqrt(absoluteError)/(n*m)
        print *,"The error of the gathered solution is", absoluteError
    end if

    call MPI_FINALIZE(ierr)
end program jacobi
